from __future__ import annotations
import asyncio
import logging
import time
from typing import Any

import httpx

from gramix.constants import (
    API_BASE_URL,
    API_FILE_URL,
    DEFAULT_TIMEOUT,
    MAX_MESSAGE_LENGTH,
    RETRY_ATTEMPTS,
    RETRY_DELAY,
    RETRY_BACKOFF,
    TOKEN_ENV_KEY,
    _SENTINEL,
)
from gramix.env import get_token
from gramix.exceptions import FileError, NetworkError, RetryAfterError, TelegramAPIError, TokenError
from gramix.types.keyboard import Inline, Reply, RemoveKeyboard
from gramix.types.message import Message
from gramix.types.user import User

logger = logging.getLogger(__name__)


class Bot:
    def __init__(
        self,
        token: str | None = None,
        *,
        timeout: float = DEFAULT_TIMEOUT,
        parse_mode: str | None = None,
    ) -> None:
        resolved_token = token or get_token()
        if not resolved_token:
            raise TokenError(
                f"Токен не передан и не найден в '{TOKEN_ENV_KEY}'. "
                "Получи токен у @BotFather."
            )
        self._token = resolved_token
        self._timeout = timeout
        self._default_parse_mode = parse_mode
        self._me: User | None = None
        self._sync_client: httpx.Client | None = None
        self._async_client: httpx.AsyncClient | None = None

    def __repr__(self) -> str:
        hint = f"{self._token[:8]}...{self._token[-4:]}"
        return f"<Bot token={hint}>"

    def _build_url(self, method: str) -> str:
        return API_BASE_URL.format(token=self._token, method=method)

    def _get_sync_client(self) -> httpx.Client:
        if self._sync_client is None or self._sync_client.is_closed:
            self._sync_client = httpx.Client(timeout=self._timeout)
        return self._sync_client

    def _parse_response(self, response: httpx.Response) -> Any:
        data = response.json()
        if not data.get("ok"):
            code = data.get("error_code", 0)
            description = data.get("description", "Unknown error")
            if code == 429:
                retry_after = data.get("parameters", {}).get("retry_after", 30)
                raise RetryAfterError(retry_after)
            raise TelegramAPIError(code, description)
        return data["result"]

    def _request(self, method: str, payload: dict) -> Any:
        url = self._build_url(method)
        clean_payload = {k: v for k, v in payload.items() if v is not None}
        delay = RETRY_DELAY
        last_retry_after: RetryAfterError | None = None

        for attempt in range(1, RETRY_ATTEMPTS + 1):
            try:
                response = self._get_sync_client().post(url, json=clean_payload)
                return self._parse_response(response)
            except RetryAfterError as e:
                last_retry_after = e
                logger.warning("Flood control: ждём %ds.", e.retry_after)
                time.sleep(e.retry_after)
            except TelegramAPIError:
                raise
            except httpx.TransportError as e:
                if attempt == RETRY_ATTEMPTS:
                    raise NetworkError(f"Сетевая ошибка: {e}") from e
                logger.warning("Сетевая ошибка (попытка %d/%d): %s", attempt, RETRY_ATTEMPTS, e)
                time.sleep(delay)
                delay *= RETRY_BACKOFF

        if last_retry_after is not None:
            raise last_retry_after
        raise NetworkError(f"Запрос {method} не выполнен после {RETRY_ATTEMPTS} попыток.")

    async def _async_request(self, method: str, payload: dict) -> Any:
        url = self._build_url(method)
        clean_payload = {k: v for k, v in payload.items() if v is not None}
        delay = RETRY_DELAY
        last_retry_after: RetryAfterError | None = None

        if self._async_client is None or self._async_client.is_closed:
            self._async_client = httpx.AsyncClient(timeout=self._timeout)

        for attempt in range(1, RETRY_ATTEMPTS + 1):
            try:
                response = await self._async_client.post(url, json=clean_payload)
                return self._parse_response(response)
            except RetryAfterError as e:
                last_retry_after = e
                logger.warning("Flood control: ждём %ds.", e.retry_after)
                await asyncio.sleep(e.retry_after)
            except TelegramAPIError:
                raise
            except httpx.TransportError as e:
                if attempt == RETRY_ATTEMPTS:
                    raise NetworkError(f"Сетевая ошибка: {e}") from e
                logger.warning("Сетевая ошибка (попытка %d/%d): %s", attempt, RETRY_ATTEMPTS, e)
                await asyncio.sleep(delay)
                delay *= RETRY_BACKOFF

        if last_retry_after is not None:
            raise last_retry_after
        raise NetworkError(f"Запрос {method} не выполнен после {RETRY_ATTEMPTS} попыток.")

    def _keyboard_dict(self, keyboard: Inline | Reply | RemoveKeyboard | None) -> dict | None:
        return keyboard.to_dict() if keyboard is not None else None

    def _effective_parse_mode(self, parse_mode: Any) -> str | None:
        return self._default_parse_mode if parse_mode is _SENTINEL else parse_mode

    def get_me(self) -> User:
        if self._me is None:
            self._me = User.from_dict(self._request("getMe", {}))
        return self._me

    def refresh_me(self) -> User:
        """Принудительно обновить кэш данных бота (имя/username можно менять через BotFather)."""
        self._me = User.from_dict(self._request("getMe", {}))
        return self._me

    async def async_get_me(self) -> User:
        if self._me is None:
            self._me = User.from_dict(await self._async_request("getMe", {}))
        return self._me

    async def async_refresh_me(self) -> User:
        """Принудительно обновить кэш данных бота (async-версия)."""
        self._me = User.from_dict(await self._async_request("getMe", {}))
        return self._me

    def send_message(
        self,
        chat_id: int | str,
        text: str,
        *,
        reply_to_message_id: int | None = None,
        keyboard: Inline | Reply | RemoveKeyboard | None = None,
        parse_mode: str | None = _SENTINEL,
        disable_preview: bool = False,
        auto_split: bool = False,
    ) -> Message:
        if auto_split and len(text) > MAX_MESSAGE_LENGTH:
            chunks = [text[i:i + MAX_MESSAGE_LENGTH] for i in range(0, len(text), MAX_MESSAGE_LENGTH)]
            last: Message | None = None
            for chunk in chunks:
                last = self.send_message(
                    chat_id, chunk,
                    reply_to_message_id=reply_to_message_id,
                    parse_mode=parse_mode,
                    disable_preview=disable_preview,
                )
            if last is None:
                raise ValueError("auto_split: текст пустой после разбивки.")
            return last

        data = self._request("sendMessage", {
            "chat_id": chat_id,
            "text": text,
            "reply_to_message_id": reply_to_message_id,
            "reply_markup": self._keyboard_dict(keyboard),
            "parse_mode": self._effective_parse_mode(parse_mode),
            "disable_web_page_preview": disable_preview if disable_preview else None,
        })
        return Message.from_dict(data, self)

    async def async_send_message(
        self,
        chat_id: int | str,
        text: str,
        *,
        reply_to_message_id: int | None = None,
        keyboard: Inline | Reply | RemoveKeyboard | None = None,
        parse_mode: str | None = _SENTINEL,
        disable_preview: bool = False,
        auto_split: bool = False,
    ) -> Message:
        if auto_split and len(text) > MAX_MESSAGE_LENGTH:
            chunks = [text[i:i + MAX_MESSAGE_LENGTH] for i in range(0, len(text), MAX_MESSAGE_LENGTH)]
            last: Message | None = None
            for chunk in chunks:
                last = await self.async_send_message(
                    chat_id, chunk,
                    reply_to_message_id=reply_to_message_id,
                    parse_mode=parse_mode,
                    disable_preview=disable_preview,
                )
            if last is None:
                raise ValueError("auto_split: текст пустой после разбивки.")
            return last

        data = await self._async_request("sendMessage", {
            "chat_id": chat_id,
            "text": text,
            "reply_to_message_id": reply_to_message_id,
            "reply_markup": self._keyboard_dict(keyboard),
            "parse_mode": self._effective_parse_mode(parse_mode),
            "disable_web_page_preview": disable_preview if disable_preview else None,
        })
        return Message.from_dict(data, self)

    def forward_message(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_id: int,
        *,
        disable_notification: bool = False,
    ) -> Message:
        data = self._request("forwardMessage", {
            "chat_id": chat_id,
            "from_chat_id": from_chat_id,
            "message_id": message_id,
            "disable_notification": disable_notification if disable_notification else None,
        })
        return Message.from_dict(data, self)

    def copy_message(
        self,
        chat_id: int | str,
        from_chat_id: int | str,
        message_id: int,
        *,
        caption: str | None = None,
        parse_mode: str | None = _SENTINEL,
        keyboard: Inline | Reply | None = None,
    ) -> int:
        data = self._request("copyMessage", {
            "chat_id": chat_id,
            "from_chat_id": from_chat_id,
            "message_id": message_id,
            "caption": caption,
            "parse_mode": self._effective_parse_mode(parse_mode),
            "reply_markup": self._keyboard_dict(keyboard),
        })
        return int(data["message_id"])

    def edit_message_text(
        self,
        chat_id: int | str,
        message_id: int,
        text: str,
        *,
        keyboard: Inline | None = None,
        parse_mode: str | None = _SENTINEL,
    ) -> Message:
        data = self._request("editMessageText", {
            "chat_id": chat_id,
            "message_id": message_id,
            "text": text,
            "reply_markup": self._keyboard_dict(keyboard),
            "parse_mode": self._effective_parse_mode(parse_mode),
        })
        return Message.from_dict(data, self)

    def edit_message_keyboard(
        self,
        chat_id: int | str,
        message_id: int,
        keyboard: Inline | None,
    ) -> bool:
        return self._request("editMessageReplyMarkup", {
            "chat_id": chat_id,
            "message_id": message_id,
            "reply_markup": self._keyboard_dict(keyboard),
        })

    def delete_message(self, chat_id: int | str, message_id: int) -> bool:
        return self._request("deleteMessage", {
            "chat_id": chat_id,
            "message_id": message_id,
        })

    def pin_chat_message(
        self,
        chat_id: int | str,
        message_id: int,
        *,
        disable_notification: bool = False,
    ) -> bool:
        return self._request("pinChatMessage", {
            "chat_id": chat_id,
            "message_id": message_id,
            "disable_notification": disable_notification if disable_notification else None,
        })

    def unpin_chat_message(self, chat_id: int | str, message_id: int) -> bool:
        return self._request("unpinChatMessage", {
            "chat_id": chat_id,
            "message_id": message_id,
        })

    def set_message_reaction(
        self,
        chat_id: int | str,
        message_id: int,
        reaction: str,
        *,
        is_big: bool = False,
    ) -> bool:
        return self._request("setMessageReaction", {
            "chat_id": chat_id,
            "message_id": message_id,
            "reaction": [{"type": "emoji", "emoji": reaction}],
            "is_big": is_big if is_big else None,
        })

    def send_photo(
        self,
        chat_id: int | str,
        photo: str,
        *,
        caption: str | None = None,
        keyboard: Inline | Reply | None = None,
        parse_mode: str | None = _SENTINEL,
    ) -> Message:
        data = self._request("sendPhoto", {
            "chat_id": chat_id,
            "photo": photo,
            "caption": caption,
            "reply_markup": self._keyboard_dict(keyboard),
            "parse_mode": self._effective_parse_mode(parse_mode),
        })
        return Message.from_dict(data, self)

    def send_video(
        self,
        chat_id: int | str,
        video: str,
        *,
        caption: str | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        keyboard: Inline | Reply | None = None,
        parse_mode: str | None = _SENTINEL,
    ) -> Message:
        data = self._request("sendVideo", {
            "chat_id": chat_id,
            "video": video,
            "caption": caption,
            "duration": duration,
            "width": width,
            "height": height,
            "reply_markup": self._keyboard_dict(keyboard),
            "parse_mode": self._effective_parse_mode(parse_mode),
        })
        return Message.from_dict(data, self)

    def send_audio(
        self,
        chat_id: int | str,
        audio: str,
        *,
        caption: str | None = None,
        duration: int | None = None,
        performer: str | None = None,
        title: str | None = None,
        keyboard: Inline | Reply | None = None,
    ) -> Message:
        data = self._request("sendAudio", {
            "chat_id": chat_id,
            "audio": audio,
            "caption": caption,
            "duration": duration,
            "performer": performer,
            "title": title,
            "reply_markup": self._keyboard_dict(keyboard),
        })
        return Message.from_dict(data, self)

    def send_voice(
        self,
        chat_id: int | str,
        voice: str,
        *,
        caption: str | None = None,
        duration: int | None = None,
    ) -> Message:
        data = self._request("sendVoice", {
            "chat_id": chat_id,
            "voice": voice,
            "caption": caption,
            "duration": duration,
        })
        return Message.from_dict(data, self)

    def send_document(
        self,
        chat_id: int | str,
        document: str,
        *,
        caption: str | None = None,
        keyboard: Inline | Reply | None = None,
    ) -> Message:
        data = self._request("sendDocument", {
            "chat_id": chat_id,
            "document": document,
            "caption": caption,
            "reply_markup": self._keyboard_dict(keyboard),
        })
        return Message.from_dict(data, self)

    def send_sticker(self, chat_id: int | str, sticker: str) -> Message:
        data = self._request("sendSticker", {"chat_id": chat_id, "sticker": sticker})
        return Message.from_dict(data, self)

    def send_chat_action(self, chat_id: int | str, action: str) -> bool:
        return self._request("sendChatAction", {"chat_id": chat_id, "action": action})

    def get_file(self, file_id: str) -> dict:
        return self._request("getFile", {"file_id": file_id})

    def download_file(self, file_id: str) -> bytes:
        file_info = self.get_file(file_id)
        file_path = file_info.get("file_path")
        if not file_path:
            raise FileError("Файл не найден или недоступен для скачивания.")
        url = API_FILE_URL.format(token=self._token, path=file_path)
        response = self._get_sync_client().get(url)
        if response.status_code != 200:
            raise FileError(f"Не удалось скачать файл: HTTP {response.status_code}")
        return response.content

    def answer_callback_query(
        self,
        callback_query_id: str,
        text: str | None = None,
        *,
        show_alert: bool = False,
        url: str | None = None,
        cache_time: int = 0,
    ) -> bool:
        return self._request("answerCallbackQuery", {
            "callback_query_id": callback_query_id,
            "text": text,
            "show_alert": show_alert if show_alert else None,
            "url": url,
            "cache_time": cache_time if cache_time else None,
        })

    def answer_inline_query(
        self,
        inline_query_id: str,
        results: list[dict],
        *,
        cache_time: int = 300,
        is_personal: bool = False,
        next_offset: str | None = None,
    ) -> bool:
        return self._request("answerInlineQuery", {
            "inline_query_id": inline_query_id,
            "results": results,
            "cache_time": cache_time,
            "is_personal": is_personal if is_personal else None,
            "next_offset": next_offset,
        })

    def get_chat(self, chat_id: int | str) -> dict:
        return self._request("getChat", {"chat_id": chat_id})

    def get_chat_member(self, chat_id: int | str, user_id: int) -> dict:
        return self._request("getChatMember", {"chat_id": chat_id, "user_id": user_id})

    def get_chat_members_count(self, chat_id: int | str) -> int:
        return self._request("getChatMemberCount", {"chat_id": chat_id})

    def ban_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        *,
        until_date: int | None = None,
    ) -> bool:
        return self._request("banChatMember", {
            "chat_id": chat_id,
            "user_id": user_id,
            "until_date": until_date,
        })

    def unban_chat_member(self, chat_id: int | str, user_id: int) -> bool:
        return self._request("unbanChatMember", {
            "chat_id": chat_id,
            "user_id": user_id,
            "only_if_banned": True,
        })

    def restrict_chat_member(
        self,
        chat_id: int | str,
        user_id: int,
        permissions: dict,
        *,
        until_date: int | None = None,
    ) -> bool:
        return self._request("restrictChatMember", {
            "chat_id": chat_id,
            "user_id": user_id,
            "permissions": permissions,
            "until_date": until_date,
        })

    def leave_chat(self, chat_id: int | str) -> bool:
        return self._request("leaveChat", {"chat_id": chat_id})

    def set_my_commands(
        self,
        commands: list[dict],
        *,
        scope: dict | None = None,
    ) -> bool:
        return self._request("setMyCommands", {
            "commands": commands,
            "scope": scope,
        })

    def delete_my_commands(self) -> bool:
        return self._request("deleteMyCommands", {})

    def get_updates(
        self,
        offset: int | None = None,
        limit: int = 100,
        timeout: int = 30,
        allowed_updates: list[str] | None = None,
    ) -> list[dict]:
        return self._request("getUpdates", {
            "offset": offset,
            "limit": limit,
            "timeout": timeout,
            "allowed_updates": allowed_updates,
        })

    async def async_get_updates(
        self,
        offset: int | None = None,
        limit: int = 100,
        timeout: int = 30,
        allowed_updates: list[str] | None = None,
    ) -> list[dict]:
        return await self._async_request("getUpdates", {
            "offset": offset,
            "limit": limit,
            "timeout": timeout,
            "allowed_updates": allowed_updates,
        })

    def set_webhook(
        self,
        url: str,
        *,
        secret_token: str | None = None,
        max_connections: int = 40,
        allowed_updates: list[str] | None = None,
    ) -> bool:
        return self._request("setWebhook", {
            "url": url,
            "secret_token": secret_token,
            "max_connections": max_connections,
            "allowed_updates": allowed_updates,
        })

    def delete_webhook(self) -> bool:
        return self._request("deleteWebhook", {})

    def get_webhook_info(self) -> dict:
        return self._request("getWebhookInfo", {})

    def close(self) -> None:
        if self._sync_client and not self._sync_client.is_closed:
            self._sync_client.close()

    async def async_close(self) -> None:
        if self._async_client and not self._async_client.is_closed:
            await self._async_client.aclose()



API_BASE_URL: str = "https://api.telegram.org/bot{token}/{method}"
API_FILE_URL: str = "https://api.telegram.org/file/bot{token}/{path}"

# Shared sentinel — единственный экземпляр, используется в bot.py и types/message.py
# для различия «пользователь передал None» от «пользователь ничего не передал».
_SENTINEL: object = object()

DEFAULT_TIMEOUT: float = 35.0  # POLLING_TIMEOUT + 10с буфер чтобы httpx не срубил соединение
POLLING_TIMEOUT: int = 25      # Рекомендовано Telegram: 20–30с; было 5с → ~12 req/min впустую
RETRY_ATTEMPTS: int = 3
RETRY_DELAY: float = 1.0
RETRY_BACKOFF: float = 2.0

MAX_MESSAGE_LENGTH: int = 4096
MAX_CAPTION_LENGTH: int = 1024
MAX_CALLBACK_DATA_LENGTH: int = 64

TOKEN_ENV_KEY: str = "BOT_TOKEN"
DEBUG_ENV_KEY: str = "DEBUG"
WEBHOOK_URL_ENV_KEY: str = "WEBHOOK_URL"


class ParseMode:
    HTML = "HTML"
    MARKDOWN = "MarkdownV2"
    MARKDOWN_LEGACY = "Markdown"



from __future__ import annotations
import asyncio
import logging
import threading
import time
from collections.abc import Callable
from datetime import datetime
from typing import Any

from gramix.bot import Bot
from gramix.constants import POLLING_TIMEOUT
from gramix.exceptions import NetworkError, TelegramAPIError, WebhookError
from gramix.middleware import MiddlewareManager
from gramix.router import Router
from gramix.types.callback import CallbackQuery
from gramix.types.chat_member import ChatMemberUpdated
from gramix.types.inline_query import InlineQuery
from gramix.types.message import Message

logger = logging.getLogger(__name__)


class Dispatcher:
    def __init__(self, bot: Bot) -> None:
        self._bot = bot
        self._routers: list[Router] = []
        self._middleware = MiddlewareManager()
        self._running = False
        self._startup_handlers: list[Callable] = []
        self._shutdown_handlers: list[Callable] = []

    def include(self, router: Router) -> None:
        self._routers.append(router)

    @property
    def middleware(self) -> Callable:
        return self._middleware.register

    def on_startup(self, func: Callable) -> Callable:
        self._startup_handlers.append(func)
        return func

    def on_shutdown(self, func: Callable) -> Callable:
        self._shutdown_handlers.append(func)
        return func

    def _dispatch(self, raw_update: dict) -> None:
        if "message" in raw_update:
            msg = Message.from_dict(raw_update["message"], self._bot)
            if self._middleware._middlewares:
                self._middleware.run(msg, self._route_message)
            else:
                self._route_message(msg)

        elif "edited_message" in raw_update:
            msg = Message.from_dict(raw_update["edited_message"], self._bot)
            if self._middleware._middlewares:
                self._middleware.run(msg, self._route_message)
            else:
                self._route_message(msg)

        elif "channel_post" in raw_update:
            msg = Message.from_dict(raw_update["channel_post"], self._bot)
            if self._middleware._middlewares:
                self._middleware.run(msg, self._route_message)
            else:
                self._route_message(msg)

        elif "edited_channel_post" in raw_update:
            msg = Message.from_dict(raw_update["edited_channel_post"], self._bot)
            if self._middleware._middlewares:
                self._middleware.run(msg, self._route_message)
            else:
                self._route_message(msg)

        elif "callback_query" in raw_update:
            cb = CallbackQuery.from_dict(raw_update["callback_query"], self._bot)
            for router in self._routers:
                if router.process_callback(cb):
                    break

        elif "inline_query" in raw_update:
            query = InlineQuery.from_dict(raw_update["inline_query"], self._bot)
            for router in self._routers:
                if router.process_inline(query):
                    break

        elif "my_chat_member" in raw_update or "chat_member" in raw_update:
            key = "my_chat_member" if "my_chat_member" in raw_update else "chat_member"
            update = ChatMemberUpdated.from_dict(raw_update[key])
            for router in self._routers:
                router.process_chat_member(update)

    async def _async_dispatch(self, raw_update: dict) -> None:
        if "message" in raw_update:
            msg = Message.from_dict(raw_update["message"], self._bot)
            if self._middleware._middlewares:
                await self._middleware.async_run(msg, self._async_route_message)
            else:
                await self._async_route_message(msg)

        elif "edited_message" in raw_update:
            msg = Message.from_dict(raw_update["edited_message"], self._bot)
            if self._middleware._middlewares:
                await self._middleware.async_run(msg, self._async_route_message)
            else:
                await self._async_route_message(msg)

        elif "channel_post" in raw_update:
            msg = Message.from_dict(raw_update["channel_post"], self._bot)
            if self._middleware._middlewares:
                await self._middleware.async_run(msg, self._async_route_message)
            else:
                await self._async_route_message(msg)

        elif "edited_channel_post" in raw_update:
            msg = Message.from_dict(raw_update["edited_channel_post"], self._bot)
            if self._middleware._middlewares:
                await self._middleware.async_run(msg, self._async_route_message)
            else:
                await self._async_route_message(msg)

        elif "callback_query" in raw_update:
            cb = CallbackQuery.from_dict(raw_update["callback_query"], self._bot)
            for router in self._routers:
                if await router.async_process_callback(cb):
                    break

        elif "inline_query" in raw_update:
            query = InlineQuery.from_dict(raw_update["inline_query"], self._bot)
            for router in self._routers:
                if await router.async_process_inline(query):
                    break

        elif "my_chat_member" in raw_update or "chat_member" in raw_update:
            key = "my_chat_member" if "my_chat_member" in raw_update else "chat_member"
            update = ChatMemberUpdated.from_dict(raw_update[key])
            for router in self._routers:
                await router.async_process_chat_member(update)

    def _route_message(self, msg: Message) -> None:
        for router in self._routers:
            if router.process_message(msg):
                break

    async def _async_route_message(self, msg: Message) -> None:
        for router in self._routers:
            if await router.async_process_message(msg):
                break

    def run(
        self,
        *,
        webhook: bool = False,
        host: str = "0.0.0.0",
        port: int = 8080,
        webhook_url: str | None = None,
        backend: str = "raw",
        drop_pending: bool = True,
    ) -> None:
        if webhook:
            self._run_webhook(host=host, port=port, url=webhook_url, backend=backend)
        else:
            self._run_polling(drop_pending=drop_pending)

    def _print_banner(self, username: str, user_id: int, mode: str) -> None:
        from gramix import __version__
        started_at = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        logger.info("━" * 40)
        logger.info("  gramix v%s", __version__)
        logger.info("  Бот:    @%s (id=%d)", username, user_id)
        logger.info("  Режим:  %s", mode)
        logger.info("  Старт:  %s", started_at)
        logger.info("━" * 40)

    def _run_polling(self, *, drop_pending: bool = True) -> None:
        me = self._bot.get_me()
        self._print_banner(me.username or str(me.id), me.id, "polling")
        self._call_handlers(self._startup_handlers)

        if drop_pending:
            try:
                updates = self._bot.get_updates(timeout=0, limit=100)
                while updates:
                    offset = updates[-1]["update_id"] + 1
                    updates = self._bot.get_updates(offset=offset, timeout=0, limit=100)
            except Exception:
                pass

        self._running = True
        offset: int | None = None

        def poll() -> None:
            nonlocal offset
            while self._running:
                try:
                    updates = self._bot.get_updates(offset=offset, timeout=POLLING_TIMEOUT)
                    for update in updates:
                        offset = update["update_id"] + 1
                        try:
                            self._dispatch(update)
                        except Exception:
                            logger.exception("Ошибка при обработке update %d:", update["update_id"])
                except NetworkError as e:
                    if self._running:
                        logger.warning("Сетевая ошибка: %s. Переподключение...", e)
                        time.sleep(3)
                except TelegramAPIError as e:
                    logger.error("Telegram API ошибка: %s", e)
                    time.sleep(1)
                except Exception:
                    if self._running:
                        logger.exception("Неожиданная ошибка в polling:")
                        time.sleep(1)

        import signal as _signal

        def _stop(*_: Any) -> None:
            logger.info("Остановка бота...")
            self._running = False

        try:
            _signal.signal(_signal.SIGINT, _stop)
            _signal.signal(_signal.SIGTERM, _stop)
        except (OSError, ValueError):
            pass

        thread = threading.Thread(target=poll, daemon=True)
        thread.start()

        logger.info("Polling запущен. Ctrl+C для остановки.")
        try:
            while thread.is_alive():
                thread.join(timeout=0.2)
        except (KeyboardInterrupt, SystemExit):
            logger.info("Остановка бота...")
            self._running = False
            self._bot.close()
            thread.join(timeout=2)
        finally:
            self._call_handlers(self._shutdown_handlers)
            logger.info("Бот остановлен.")

    async def _async_polling(self, *, drop_pending: bool = True) -> None:
        me = await self._bot.async_get_me()
        self._print_banner(me.username or str(me.id), me.id, "async polling")
        await self._async_call_handlers(self._startup_handlers)

        if drop_pending:
            try:
                updates = await self._bot.async_get_updates(timeout=0, limit=100)
                while updates:
                    offset = updates[-1]["update_id"] + 1
                    updates = await self._bot.async_get_updates(offset=offset, timeout=0, limit=100)
            except Exception:
                pass

        self._running = True
        offset: int | None = None

        logger.info("Async polling запущен. Ctrl+C для остановки.")
        try:
            while self._running:
                try:
                    updates = await self._bot.async_get_updates(
                        offset=offset, timeout=POLLING_TIMEOUT
                    )
                    for update in updates:
                        offset = update["update_id"] + 1
                        try:
                            await self._async_dispatch(update)
                        except Exception:
                            logger.exception(
                                "Ошибка при обработке update %d:", update["update_id"]
                            )
                except NetworkError as e:
                    logger.warning("Сетевая ошибка: %s.", e)
                    await asyncio.sleep(3)
                except TelegramAPIError as e:
                    logger.error("Telegram API ошибка: %s", e)
                    await asyncio.sleep(1)
                except asyncio.CancelledError:
                    break
                except Exception:
                    logger.exception("Неожиданная ошибка:")
                    await asyncio.sleep(1)
        finally:
            await self._async_call_handlers(self._shutdown_handlers)
            await self._bot.async_close()
            logger.info("Бот остановлен.")

    def run_async(self, *, drop_pending: bool = True) -> None:
        asyncio.run(self._async_polling(drop_pending=drop_pending))

    def _run_webhook(self, host: str, port: int, url: str | None, backend: str = "raw") -> None:
        from gramix.env import get_webhook_url

        webhook_url = url or get_webhook_url()
        if not webhook_url:
            raise WebhookError(
                "URL webhook не указан. Передай параметр url= или установи WEBHOOK_URL в .env."
            )

        self._bot.set_webhook(webhook_url)
        me = self._bot.get_me()
        self._print_banner(me.username or str(me.id), me.id, f"webhook [{backend}] {webhook_url}")
        logger.info("Сервер: %s:%d", host, port)
        self._call_handlers(self._startup_handlers)

        if backend == "aiohttp":
            self._run_webhook_aiohttp(host=host, port=port)
        elif backend == "fastapi":
            self._run_webhook_fastapi(host=host, port=port)
        else:
            self._run_webhook_raw(host=host, port=port)

    def _run_webhook_raw(self, host: str, port: int) -> None:
        import socket

        self._running = True
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((host, port))
        server.listen(5)
        server.settimeout(1.0)

        try:
            while self._running:
                try:
                    conn, _ = server.accept()
                    self._handle_webhook_request(conn)
                except TimeoutError:
                    continue
        except KeyboardInterrupt:
            pass
        finally:
            server.close()
            self._bot.delete_webhook()
            self._call_handlers(self._shutdown_handlers)
            self._bot.close()
            logger.info("Webhook сервер остановлен.")

    def _run_webhook_aiohttp(self, host: str, port: int) -> None:
        try:
            from aiohttp import web
        except ImportError:
            raise WebhookError(
                "aiohttp не установлен. Установи: pip install gramix[aiohttp]"
            )

        async def handle(request: web.Request) -> web.Response:
            try:
                data = await request.json()
                await self._async_dispatch(data)
            except Exception:
                logger.exception("Ошибка при обработке webhook запроса:")
            return web.Response(text="OK")

        async def run() -> None:
            app = web.Application()
            app.router.add_post("/", handle)
            runner = web.AppRunner(app)
            await runner.setup()
            site = web.TCPSite(runner, host, port)
            await site.start()
            logger.info("aiohttp webhook сервер запущен.")
            try:
                while True:
                    await asyncio.sleep(3600)
            except (KeyboardInterrupt, asyncio.CancelledError):
                pass
            finally:
                await runner.cleanup()
                self._bot.delete_webhook()
                self._call_handlers(self._shutdown_handlers)
                self._bot.close()
                logger.info("Webhook сервер остановлен.")

        asyncio.run(run())

    def _run_webhook_fastapi(self, host: str, port: int) -> None:
        try:
            import uvicorn
            from contextlib import asynccontextmanager
            from fastapi import FastAPI, Request, Response
        except ImportError:
            raise WebhookError(
                "fastapi/uvicorn не установлены. Установи: pip install gramix[fastapi]"
            )

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            yield
            self._bot.delete_webhook()
            self._call_handlers(self._shutdown_handlers)
            self._bot.close()
            logger.info("Webhook сервер остановлен.")

        app = FastAPI(lifespan=lifespan)

        @app.post("/")
        async def handle(request: Request) -> Response:
            try:
                data = await request.json()
                await self._async_dispatch(data)
            except Exception:
                logger.exception("Ошибка при обработке webhook запроса:")
            return Response(content="OK")

        logger.info("fastapi webhook сервер запущен.")
        uvicorn.run(app, host=host, port=port)

    def _handle_webhook_request(self, conn: Any) -> None:
        import json
        try:
            raw = b""
            conn.settimeout(5.0)
            while True:
                chunk = conn.recv(8192)
                if not chunk:
                    break
                raw += chunk
                if b"\r\n\r\n" in raw:
                    header_part, _, body_part = raw.partition(b"\r\n\r\n")
                    headers = {}
                    for line in header_part.decode("utf-8", errors="replace").splitlines()[1:]:
                        if ":" in line:
                            k, _, v = line.partition(":")
                            headers[k.strip().lower()] = v.strip()
                    content_length = int(headers.get("content-length", 0))
                    while len(body_part) < content_length:
                        chunk = conn.recv(8192)
                        if not chunk:
                            break
                        body_part += chunk
                    if body_part:
                        self._dispatch(json.loads(body_part.decode("utf-8")))
                    break

            conn.send(b"HTTP/1.1 200 OK\r\nContent-Length: 2\r\nContent-Type: text/plain\r\n\r\nOK")
        except Exception:
            logger.exception("Ошибка при обработке webhook запроса:")
        finally:
            conn.close()

    def _call_handlers(self, handlers: list[Callable]) -> None:
        for handler in handlers:
            try:
                if asyncio.iscoroutinefunction(handler):
                    logger.warning(
                        "Lifecycle handler '%s' is async but running in sync mode — "
                        "используй run_async() или объяви handler как sync-функцию.",
                        handler.__name__,
                    )
                    asyncio.get_event_loop().run_until_complete(handler())
                else:
                    handler()
            except Exception:
                logger.exception("Ошибка в lifecycle handler:")

    async def _async_call_handlers(self, handlers: list[Callable]) -> None:
        for handler in handlers:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler()
                else:
                    handler()
            except Exception:
                logger.exception("Ошибка в lifecycle handler:")



from __future__ import annotations
import json
import logging
import sqlite3
import threading
from abc import ABC, abstractmethod
from typing import Any

from gramix.exceptions import FSMError

logger = logging.getLogger(__name__)

_MAX_MEMORY_SIZE = 10_000


class Step:
    _name: str
    _owner: str

    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name
        self._owner = owner.__name__

    def __repr__(self) -> str:
        return f"{self._owner}.{self._name}"


class StateMeta(type):
    def __new__(mcs, name: str, bases: tuple, namespace: dict) -> StateMeta:
        steps = [key for key, value in namespace.items() if isinstance(value, Step)]
        namespace["_steps"] = steps
        return super().__new__(mcs, name, bases, namespace)


class State(metaclass=StateMeta):
    _steps: list[str] = []

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        _register_state(cls)


class StateContext:
    __slots__ = ("_state_class", "_current_step", "data", "_user_id", "_storage")

    def __init__(
        self,
        user_id: int,
        storage: BaseStorage,
        state_class: type[State] | None = None,
        current_step: str | None = None,
        data: dict[str, Any] | None = None,
    ) -> None:
        self._user_id = user_id
        self._storage = storage
        self._state_class = state_class
        self._current_step = current_step
        self.data: dict[str, Any] = data or {}

    @property
    def is_active(self) -> bool:
        return self._state_class is not None

    @property
    def current(self) -> str | None:
        if self._state_class and self._current_step:
            return f"{self._state_class.__name__}.{self._current_step}"
        return None

    def set(self, step: Step) -> None:
        if not hasattr(step, "_owner") or not hasattr(step, "_name"):
            raise FSMError("Step не инициализирован. Используй его как атрибут класса State.")
        state_class = _find_state_class(step._owner)
        if state_class is None:
            raise FSMError(f"Класс состояния '{step._owner}' не найден.")
        if step._name not in state_class._steps:
            raise FSMError(f"Шаг '{step._name}' не найден в '{step._owner}'.")
        self._state_class = state_class
        self._current_step = step._name
        self._storage.set(self._user_id, step._owner, step._name, self.data)
        logger.debug("FSM [%d]: → %s", self._user_id, self.current)

    def next(self) -> bool:
        if not self._state_class or not self._current_step:
            raise FSMError("Состояние не установлено.")
        steps = self._state_class._steps
        idx = steps.index(self._current_step)
        if idx + 1 >= len(steps):
            self.finish()
            return False
        self._current_step = steps[idx + 1]
        self._storage.set(self._user_id, self._state_class.__name__, self._current_step, self.data)
        logger.debug("FSM [%d]: → %s", self._user_id, self.current)
        return True

    def prev(self) -> bool:
        if not self._state_class or not self._current_step:
            raise FSMError("Состояние не установлено.")
        steps = self._state_class._steps
        idx = steps.index(self._current_step)
        if idx == 0:
            return False
        self._current_step = steps[idx - 1]
        self._storage.set(self._user_id, self._state_class.__name__, self._current_step, self.data)
        logger.debug("FSM [%d]: ← %s", self._user_id, self.current)
        return True

    def finish(self) -> None:
        logger.debug("FSM [%d]: finish", self._user_id)
        self._reset()
        self._storage.delete(self._user_id)

    def _reset(self) -> None:
        self._state_class = None
        self._current_step = None
        self.data.clear()

    def matches(self, step: Step) -> bool:
        if not self._state_class or not self._current_step:
            return False
        return self._state_class.__name__ == step._owner and self._current_step == step._name


class BaseStorage(ABC):
    @abstractmethod
    def get(self, user_id: int) -> StateContext:
        ...

    @abstractmethod
    def set(self, user_id: int, state_class: str, step: str, data: dict) -> None:
        ...

    @abstractmethod
    def delete(self, user_id: int) -> None:
        ...

    @abstractmethod
    def clear_all(self) -> None:
        ...


class MemoryStorage(BaseStorage):
    def __init__(self) -> None:
        self._storage: dict[int, StateContext] = {}

    def get(self, user_id: int) -> StateContext:
        if user_id not in self._storage:
            if len(self._storage) >= _MAX_MEMORY_SIZE:
                self._evict_inactive()
            self._storage[user_id] = StateContext(user_id, self)
        return self._storage[user_id]

    def set(self, user_id: int, state_class: str, step: str, data: dict) -> None:
        # No-op: state is held by StateContext object directly in self._storage.
        # StateContext mutates its own fields (_state_class, _current_step, data),
        # so there is nothing to persist separately for in-memory storage.
        pass

    def delete(self, user_id: int) -> None:
        if user_id in self._storage:
            self._storage[user_id]._reset()

    def clear_all(self) -> None:
        self._storage.clear()

    def _evict_inactive(self) -> None:
        inactive = [uid for uid, ctx in self._storage.items() if not ctx.is_active]
        for uid in inactive:
            del self._storage[uid]


FSMStorage = MemoryStorage


class SQLiteStorage(BaseStorage):
    def __init__(self, path: str = "gramix_fsm.db") -> None:
        self._path = path
        self._lock = threading.Lock()
        self._conn = sqlite3.connect(path, check_same_thread=False)
        self._conn.row_factory = sqlite3.Row
        self._init_db()

    def _init_db(self) -> None:
        with self._lock:
            self._conn.execute("""
                CREATE TABLE IF NOT EXISTS fsm_states (
                    user_id     INTEGER PRIMARY KEY,
                    state_class TEXT NOT NULL,
                    step        TEXT NOT NULL,
                    data        TEXT NOT NULL DEFAULT '{}'
                )
            """)
            self._conn.commit()

    def get(self, user_id: int) -> StateContext:
        with self._lock:
            row = self._conn.execute(
                "SELECT state_class, step, data FROM fsm_states WHERE user_id = ?",
                (user_id,),
            ).fetchone()

        if row is None:
            return StateContext(user_id, self)

        state_class = _find_state_class(row["state_class"])
        if state_class is None:
            return StateContext(user_id, self)

        try:
            data = json.loads(row["data"])
        except (json.JSONDecodeError, TypeError):
            data = {}

        return StateContext(
            user_id=user_id,
            storage=self,
            state_class=state_class,
            current_step=row["step"],
            data=data,
        )

    def set(self, user_id: int, state_class: str, step: str, data: dict) -> None:
        with self._lock:
            self._conn.execute("""
                INSERT INTO fsm_states (user_id, state_class, step, data)
                VALUES (?, ?, ?, ?)
                ON CONFLICT(user_id) DO UPDATE SET
                    state_class = excluded.state_class,
                    step        = excluded.step,
                    data        = excluded.data
            """, (user_id, state_class, step, json.dumps(data, ensure_ascii=False)))
            self._conn.commit()

    def delete(self, user_id: int) -> None:
        with self._lock:
            self._conn.execute("DELETE FROM fsm_states WHERE user_id = ?", (user_id,))
            self._conn.commit()

    def clear_all(self) -> None:
        with self._lock:
            self._conn.execute("DELETE FROM fsm_states")
            self._conn.commit()


_state_registry: dict[str, type[State]] = {}


def _register_state(cls: type[State]) -> None:
    _state_registry[cls.__name__] = cls


def _find_state_class(name: str) -> type[State] | None:
    return _state_registry.get(name)







from __future__ import annotations
import logging
import os
from pathlib import Path

from gramix.constants import TOKEN_ENV_KEY, DEBUG_ENV_KEY, WEBHOOK_URL_ENV_KEY

logger = logging.getLogger(__name__)


def load_env(path: str | Path = ".env") -> None:
    env_path = Path(path)
    if not env_path.exists():
        logger.debug(".env не найден: %s", env_path.resolve())
        return

    with env_path.open(encoding="utf-8") as f:
        for line_number, line in enumerate(f, start=1):
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            if "=" not in line:
                logger.warning(".env строка %d пропущена (нет '='): %s", line_number, line)
                continue
            key, _, value = line.partition("=")
            key = key.strip()
            value = value.strip().strip('"').strip("'")
            if key and key not in os.environ:
                os.environ[key] = value

    logger.debug(".env загружен: %s", env_path.resolve())


def get_token() -> str | None:
    return os.environ.get(TOKEN_ENV_KEY)


def is_debug() -> bool:
    return os.environ.get(DEBUG_ENV_KEY, "").lower() in ("1", "true", "yes")


def get_webhook_url() -> str | None:
    return os.environ.get(WEBHOOK_URL_ENV_KEY)



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

from gramix.types.user import User

if TYPE_CHECKING:
    from gramix.bot import Bot


@dataclass(slots=True)
class InlineQueryResultArticle:
    id: str
    title: str
    message_text: str
    description: str | None = None
    parse_mode: str | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "article",
            "id": self.id,
            "title": self.title,
            "input_message_content": {
                "message_text": self.message_text,
            },
        }
        if self.description:
            result["description"] = self.description
        if self.parse_mode:
            result["input_message_content"]["parse_mode"] = self.parse_mode
        return result


@dataclass(slots=True)
class InlineQueryResultPhoto:
    id: str
    photo_url: str
    thumb_url: str
    title: str | None = None
    description: str | None = None
    caption: str | None = None
    parse_mode: str | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "photo",
            "id": self.id,
            "photo_url": self.photo_url,
            "thumbnail_url": self.thumb_url,
        }
        if self.title:
            result["title"] = self.title
        if self.description:
            result["description"] = self.description
        if self.caption:
            result["caption"] = self.caption
        if self.parse_mode:
            result["parse_mode"] = self.parse_mode
        return result


@dataclass(slots=True)
class InlineQueryResultGif:
    id: str
    gif_url: str
    thumb_url: str
    title: str | None = None
    caption: str | None = None
    parse_mode: str | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "gif",
            "id": self.id,
            "gif_url": self.gif_url,
            "thumbnail_url": self.thumb_url,
        }
        if self.title:
            result["title"] = self.title
        if self.caption:
            result["caption"] = self.caption
        if self.parse_mode:
            result["parse_mode"] = self.parse_mode
        return result


@dataclass(slots=True)
class InlineQueryResultVideo:
    id: str
    video_url: str
    mime_type: str
    thumb_url: str
    title: str
    caption: str | None = None
    parse_mode: str | None = None
    description: str | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "video",
            "id": self.id,
            "video_url": self.video_url,
            "mime_type": self.mime_type,
            "thumbnail_url": self.thumb_url,
            "title": self.title,
        }
        if self.caption:
            result["caption"] = self.caption
        if self.parse_mode:
            result["parse_mode"] = self.parse_mode
        if self.description:
            result["description"] = self.description
        return result


@dataclass(slots=True)
class InlineQueryResultDocument:
    id: str
    title: str
    document_url: str
    mime_type: str
    caption: str | None = None
    parse_mode: str | None = None
    description: str | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "document",
            "id": self.id,
            "title": self.title,
            "document_url": self.document_url,
            "mime_type": self.mime_type,
        }
        if self.caption:
            result["caption"] = self.caption
        if self.parse_mode:
            result["parse_mode"] = self.parse_mode
        if self.description:
            result["description"] = self.description
        return result


@dataclass(slots=True)
class InlineQueryResultAudio:
    id: str
    audio_url: str
    title: str
    caption: str | None = None
    parse_mode: str | None = None
    performer: str | None = None
    audio_duration: int | None = None

    def to_dict(self) -> dict:
        result: dict = {
            "type": "audio",
            "id": self.id,
            "audio_url": self.audio_url,
            "title": self.title,
        }
        if self.caption:
            result["caption"] = self.caption
        if self.parse_mode:
            result["parse_mode"] = self.parse_mode
        if self.performer:
            result["performer"] = self.performer
        if self.audio_duration:
            result["audio_duration"] = self.audio_duration
        return result


class InlineQuery:
    __slots__ = ("id", "from_user", "query", "offset", "_bot")

    def __init__(
        self,
        id: str,
        from_user: User,
        query: str,
        offset: str,
        bot: Bot,
    ) -> None:
        self.id = id
        self.from_user = from_user
        self.query = query
        self.offset = offset
        self._bot = bot

    def answer(
        self,
        results: list[InlineQueryResultArticle],
        *,
        cache_time: int = 300,
        is_personal: bool = False,
        next_offset: str | None = None,
    ) -> bool:
        return self._bot.answer_inline_query(
            inline_query_id=self.id,
            results=[r.to_dict() for r in results],
            cache_time=cache_time,
            is_personal=is_personal,
            next_offset=next_offset,
        )

    @classmethod
    def from_dict(cls, data: dict, bot: Bot) -> InlineQuery:
        return cls(
            id=data["id"],
            from_user=User.from_dict(data["from"]),
            query=data.get("query", ""),
            offset=data.get("offset", ""),
            bot=bot,
        )



from __future__ import annotations
import logging

from gramix.bot import Bot
from gramix.constants import ParseMode
from gramix.dispatcher import Dispatcher
from gramix.env import load_env
from gramix.exceptions import (
    FileError,
    FilterError,
    FSMError,
    GramixError,
    MessageError,
    MiddlewareError,
    NetworkError,
    RetryAfterError,
    TelegramAPIError,
    TokenError,
    WebhookError,
)
from gramix.filters import F, CallbackPrefixFilter
from gramix.fsm import BaseStorage, FSMStorage, MemoryStorage, SQLiteStorage, State, Step
from gramix.router import Router
from gramix.types import (
    Audio,
    CallbackQuery,
    Chat,
    ChatMemberUpdated,
    ChatType,
    Document,
    Inline,
    InlineQuery,
    InlineQueryResultArticle,
    InlineQueryResultPhoto,
    InlineQueryResultGif,
    InlineQueryResultVideo,
    InlineQueryResultDocument,
    InlineQueryResultAudio,
    Message,
    PhotoSize,
    Reply,
    RemoveKeyboard,
    Sticker,
    User,
    Video,
    Voice,
)

__version__ = "0.1.3"
__author__ = "riokzy"
__license__ = "MIT"

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    "Bot",
    "Dispatcher",
    "Router",
    "ParseMode",
    "Message",
    "User",
    "Chat",
    "ChatType",
    "ChatMemberUpdated",
    "PhotoSize",
    "Document",
    "Audio",
    "Video",
    "Voice",
    "Sticker",
    "CallbackQuery",
    "InlineQuery",
    "InlineQueryResultArticle",
    "InlineQueryResultPhoto",
    "InlineQueryResultGif",
    "InlineQueryResultVideo",
    "InlineQueryResultDocument",
    "InlineQueryResultAudio",
    "Inline",
    "Reply",
    "RemoveKeyboard",
    "State",
    "Step",
    "BaseStorage",
    "MemoryStorage",
    "SQLiteStorage",
    "FSMStorage",
    "F",
    "CallbackPrefixFilter",
    "load_env",
    "GramixError",
    "TokenError",
    "TelegramAPIError",
    "NetworkError",
    "RetryAfterError",
    "MessageError",
    "FSMError",
    "FilterError",
    "MiddlewareError",
    "WebhookError",
    "FileError",
    "__version__",
]
